import cerberus
from pathlib import Path
from typing import Union


class UniqueNameChecker:
    def __init__(self):
        self.name_lookup = []

    def check(self, field, value, error):
        if value in self.name_lookup:
            error(
                field,
                f"The name field must be unique. The name {value} has already been used.",
            )
        else:
            self.name_lookup.append(value)


class RequiresExistChecker:
    def __init__(self, document):
        self.doc = document

    def check(self, field, requires, error):
        builds = self.doc["builds"]
        for value in requires:
            for build in builds:
                if value == build["name"]:
                    break
            else:
                error(field, f"{value} does not match any build name. Check spelling.")


class PathChecker:
    def __init__(self, project_dir):
        self.project_dir = project_dir

    def check(self, field, paths, error):
        paths = [(self.project_dir / the_path).resolve() for the_path in paths]

        for directory in paths:
            # Remember paths can now be directories or specific paths to files.
            if not directory.exists():
                error(field, f"{str(directory)} does not exist.")
                break


class AimCustomValidator(cerberus.Validator):
    def _check_with_output_naming_convention(self, field, value: Union[str, list]):
        # if you need more context then you can get it using the line below.
        # if self.document["buildRule"] in ["staticlib", "dynamiclib"]:

        # TODO: should we also check that the names are camelCase?
        # TODO: check outputNames are unique to prevent dependency cycle.

        def check_convention(_field, _value):
            errors = []
            if _value.startswith("lib"):
                error_str = "You should not prefix names with 'lib'."
                errors.append(error_str)

            suffix = Path(_value).suffix
            if suffix:
                error_str = f"You should not specify the suffix."
                errors.append(error_str)

            return errors

        # Bit of a hack so strings go through the same code path as lists.
        if isinstance(value, str):
            value = [value]

        for item in value:
            errors = check_convention(field, item)

            if errors:
                plural = ""
                if len(errors) > 1:
                    plural = "s"

                error_str = f"Naming convention error{plural}: {item}. " + " ".join(
                    errors
                )
                self._error(field, error_str)


def target_schema(document, project_dir):
    unique_name_checker = UniqueNameChecker()
    requires_exist_checker = RequiresExistChecker(document)
    path_checker = PathChecker(project_dir)

    schema = {
        "compiler": {"required": True, "type": "string"},
        "ar": {"required": True, "type": "string"},
        "compilerFrontend": {
            "required": True,
            "type": "string",
            "allowed": ["msvc", "gcc", "osx"],
        },
        "flags": {"type": "list", "schema": {"type": "string"}, "empty": False},
        "defines": {"type": "list", "schema": {"type": "string"}, "empty": False},
        "projectRoot": {"required": True, "type": "string", "empty": False},
        "builds": {
            "required": True,
            "type": "list",
            "schema": {
                "type": "dict",
                "schema": {
                    "name": {
                        "required": True,
                        "type": "string",
                        "check_with": unique_name_checker.check,
                    },
                    "compiler": {"required": False, "type": "string"},
                    "defines": {
                        "type": "list",
                        "schema": {"type": "string"},
                        "default": [],
                    },
                    "flags": {
                        "type": "list",
                        "schema": {"type": "string"},
                        "empty": False,
                    },
                    "requires": {
                        "type": "list",
                        "empty": False,
                        "schema": {"type": "string"},
                        "check_with": requires_exist_checker.check,
                    },
                    "buildRule": {
                        "required": True,
                        "type": "string",
                        "allowed": ["exe", "staticlib", "dynamiclib"],
                    },
                    "outputName": {
                        "required": True,
                        "type": "string",
                        "check_with": "output_naming_convention",
                    },
                    "srcDirs": {
                        "required": True,
                        "empty": False,
                        "type": "list",
                        "schema": {"type": "string"},
                        "check_with": path_checker.check,
                    },
                    "includePaths": {
                        "type": "list",
                        "empty": False,
                        "schema": {"type": "string"},
                        "check_with": path_checker.check,
                    },
                    "libraryPaths": {
                        "type": "list",
                        "empty": False,
                        "schema": {"type": "string"},
                        # you can't check the library dirs as they may not exist if the project not built before.
                        # "check_with": path_checker.check,
                        "dependencies": {"buildRule": ["exe", "dynamiclib"]},
                    },
                    "libraries": {
                        "type": "list",
                        "empty": False,
                        "schema": {"type": "string"},
                        "dependencies": {"buildRule": ["exe", "dynamiclib"]},
                        "check_with": "output_naming_convention",
                    },
                },
            },
        },
    }

    validator = AimCustomValidator()
    validator.validate(document, schema)

    # TODO: Handle schema errors. https://docs.python-cerberus.org/en/stable/errors.html
    if validator.errors:
        raise RuntimeError(validator.errors)



import functools
from typing import Dict
from aim_build.gccbuildrules import *
from aim_build.utils import *

PrefixIncludePath = functools.partial(prefix, "-I")
PrefixLibraryPath = functools.partial(prefix, "-L")
PrefixLibrary = functools.partial(prefix, "-l")
ToObjectFiles = src_to_o

FileExtensions = ["*.cpp", "*.cc", ".c"]


def get_src_files(build):
    directory = build["directory"]
    srcs = prepend_paths(directory, build["srcDirs"])
    src_dirs = [path for path in srcs if path.is_dir()]
    explicit_src_files = [path for path in srcs if path.is_file()]
    src_files = []
    for glob_pattern in FileExtensions:
        glob_files = flatten(glob(glob_pattern, src_dirs))
        src_files += glob_files

    src_files += explicit_src_files
    assert src_files, f"Fail to find any source files in {to_str(src_dirs)}."
    return src_files


def get_include_paths(build):
    directory = build["directory"]
    include_paths = build.get("includePaths", [])
    includes = prepend_paths(directory, include_paths)
    includes = PrefixIncludePath(includes)
    return includes


def get_library_paths(build):
    directory = build["directory"]
    library_paths = build.get("libraryPaths", [])
    library_paths = prepend_paths(directory, library_paths)
    library_paths = PrefixLibraryPath(library_paths)
    return library_paths


def get_library_information(build):
    libraries = build.get("libraries", [])
    link_libraries = PrefixLibrary(libraries)
    return libraries, link_libraries


def find_build(build_name, builds):
    # Note, this should never fail, as required dependencies are checked by the schema.
    for build in builds:
        if build["name"] == build_name:
            return build


class GCCBuilds:
    def add_rules(self, build):
        directory = build["build_dir"]
        ninja_path = directory / "rules.ninja"
        with ninja_path.open("w+") as ninja_file:
            writer = Writer(ninja_file)
            add_compile(writer)
            add_ar(writer)
            add_exe(writer)
            add_shared(writer)

    def build(self, build, parsed_toml, project_writer: Writer):
        the_build = build["buildRule"]

        build_name = build["name"]
        project_dir = build["directory"]
        build_dir = build["build_dir"]

        build_path = build_dir / build_name
        build_path.mkdir(parents=True, exist_ok=True)

        ninja_path = build_path / "build.ninja"
        ncompile_path = build_path / "compile.ninja"
        build["buildPath"] = build_path

        self.add_rules(build)

        with ninja_path.open("w+") as ninja_file:
            with ncompile_path.open("w+") as compile_file:
                ninja_writer = Writer(ninja_file)
                compile_writer = Writer(compile_file)
                rule_path = (build_dir / "rules.ninja").resolve()
                ninja_writer.include(escape_path(str(rule_path)))
                ninja_writer.newline()

                if the_build == "staticlib":
                    self.build_static_library(
                        ninja_writer, compile_writer, project_writer, build, parsed_toml
                    )
                elif the_build == "exe":
                    self.build_executable(
                        ninja_writer, compile_writer, project_writer, build, parsed_toml
                    )
                elif the_build == "dynamiclib":
                    self.build_dynamic_library(
                        ninja_writer, compile_writer, project_writer, build, parsed_toml
                    )
                else:
                    raise RuntimeError(f"Unknown build type {the_build}.")

    def add_compile_rule(self, nfw: Writer, build: Dict, parsed_toml):
        local_flags = build.get("flags", None)
        local_defines = build.get("defines", None)
        local_compiler = build.get("compiler", None)

        cxxflags = local_flags if local_flags else build["global_flags"]
        defines = local_defines if local_defines else build["global_defines"]
        compiler = local_compiler if local_compiler else build["global_compiler"]

        src_files = get_src_files(build)
        includes = get_include_paths(build)
        includes += self.get_required_include_information(build, parsed_toml)

        # Its very important to specify the absolute path to the obj files.
        # This prevents recompilation of files when an exe links against a library.
        # Without the absolute path to the obj files, it would build the files again
        # in the current (exe's) build location.
        build_path = build["buildPath"]
        obj_files = ToObjectFiles(src_files)
        obj_files = prepend_paths(build_path, obj_files)

        file_pairs = zip(to_str(src_files), to_str(obj_files))
        for src_file, obj_file in file_pairs:
            nfw.build(
                outputs=obj_file,
                rule="compile",
                inputs=src_file,
                variables={
                    "compiler": compiler,
                    "includes": includes,
                    "flags": cxxflags,
                    "defines": defines,
                },
            )
            nfw.newline()

        return obj_files

    def build_static_library(
        self, nfw: Writer, cfw: Writer, pfw: Writer, build: Dict, parsed_toml: Dict
    ):
        build_name = build["name"]
        library_name = self.add_static_library_naming_convention(build["outputName"])

        local_flags = build.get("flags", None)
        local_defines = build.get("defines", None)
        local_archiver = build.get("archiver", None)

        cxxflags = local_flags if local_flags else build["global_flags"]
        defines = local_defines if local_defines else build["global_defines"]
        archiver = local_archiver if local_archiver else build["global_archiver"]

        build_path = build["buildPath"]

        includes = get_include_paths(build)
        includes += self.get_required_include_information(build, parsed_toml)

        obj_files = self.add_compile_rule(cfw, build, parsed_toml)

        relative_output_name = str(build_path / library_name)

        pfw.subninja(str((Path(build_path) / "build.ninja").resolve()))

        nfw.build(
            outputs=relative_output_name,
            rule="archive",
            inputs=to_str(obj_files),
            variables={
                "archiver": archiver,
                "includes": includes,
                "flags": cxxflags,
                "defines": defines,
            },
        )
        nfw.newline()
        nfw.include(str((build_path / "compile.ninja").resolve()))

        nfw.build(rule="phony", inputs=relative_output_name, outputs=library_name)
        nfw.build(rule="phony", inputs=library_name, outputs=build_name)
        nfw.newline()

    def build_executable(
        self, nfw: Writer, cfw: Writer, pfw: Writer, build: Dict, parsed_toml: Dict
    ):
        build_name = build["name"]
        exe_name = self.add_exe_naming_convention(build["outputName"])

        local_flags = build.get("flags", None)
        local_defines = build.get("defines", None)
        local_compiler = build.get("compiler", None)

        cxxflags = local_flags if local_flags else build["global_flags"]
        defines = local_defines if local_defines else build["global_defines"]
        compiler = local_compiler if local_compiler else build["global_compiler"]

        requires = build.get("requires", [])
        build_path = build["buildPath"]

        includes = get_include_paths(build)
        includes += self.get_required_include_information(build, parsed_toml)

        library_paths = get_library_paths(build)
        (
            requires_libraries,
            requires_link_libraries,
            requires_library_paths,
            requires_library_types,
        ) = self.get_required_library_information(build, parsed_toml)
        libraries, link_libraries = get_library_information(build)

        rpath = self.get_rpath(build, parsed_toml)
        linker_args = (
            [rpath]
            + requires_library_paths
            + requires_link_libraries
            + library_paths
            + link_libraries
        )

        for requirement in requires:
            ninja_file = (build_path.parent / requirement / "build.ninja").resolve()
            assert ninja_file.exists(), f"Failed to find {str(ninja_file)}."
            nfw.subninja(escape_path(str(ninja_file)))
            nfw.newline()

        obj_files = self.add_compile_rule(cfw, build, parsed_toml)

        # Here we just need to manage the fact that the linker's library flag (-l) needs the library name without
        # lib .a/.so but the build dependency rule does need the full convention to find the build rule in the library's
        # build.ninja file.
        full_library_names = []
        for name, build_type in zip(requires_libraries, requires_library_types):
            if build_type == "staticlib":
                full_library_names.append(
                    self.add_static_library_naming_convention(name)
                )
            else:
                full_library_names.append(
                    self.add_dynamic_library_naming_convention(name)
                )

        pfw.include(str((Path(build_path) / "compile.ninja").resolve()))

        nfw.build(
            outputs=exe_name,
            rule="exe",
            inputs=to_str(obj_files),
            implicit=full_library_names,
            variables={
                "compiler": compiler,
                "includes": includes,
                "flags": cxxflags,
                "defines": defines,
                # TODO: what does exe_name do?
                "exe_name": exe_name,
                "linker_args": " ".join(linker_args),
            },
        )
        nfw.newline()
        nfw.include(str((build_path / "compile.ninja").resolve()))

        nfw.newline()
        nfw.build(rule="phony", inputs=exe_name, outputs=build_name)
        nfw.newline()

    def build_dynamic_library(
        self, nfw: Writer, cfw: Writer, pfw: Writer, build: Dict, parsed_toml: Dict
    ):
        build_name = build["name"]
        library_name = self.add_dynamic_library_naming_convention(build["outputName"])

        local_flags = build.get("flags", None)
        local_defines = build.get("defines", None)
        local_compiler = build.get("compiler", None)

        cxxflags = local_flags if local_flags else build["global_flags"]
        defines = local_defines if local_defines else build["global_defines"]
        compiler = local_compiler if local_compiler else build["global_compiler"]

        includes = get_include_paths(build)
        includes += self.get_required_include_information(build, parsed_toml)

        library_paths = get_library_paths(build)

        (
            requires_libraries,
            requires_link_libraries,
            requires_library_paths,
            requires_library_types,
        ) = self.get_required_library_information(build, parsed_toml)
        libraries, link_libraries = get_library_information(build)

        linker_args = (
            requires_link_libraries
            + requires_library_paths
            + library_paths
            + link_libraries
        )

        obj_files = self.add_compile_rule(cfw, build, parsed_toml)

        build_path = build["buildPath"]

        pfw.subninja(str((Path(build_path) / "build.ninja").resolve()))

        relative_output_name = str(build_path / library_name)
        nfw.build(
            rule="shared",
            inputs=to_str(obj_files),
            implicit=requires_libraries,
            outputs=relative_output_name,
            variables={
                "compiler": compiler,
                "includes": includes,
                "flags": " ".join(cxxflags),
                "defines": " ".join(defines),
                "lib_name": library_name,
                "linker_args": " ".join(linker_args),
            },
        )
        nfw.newline()
        nfw.include(str((build_path / "compile.ninja").resolve()))

        nfw.build(rule="phony", inputs=relative_output_name, outputs=library_name)
        nfw.build(rule="phony", inputs=library_name, outputs=build_name)
        nfw.newline()

    def get_required_library_information(self, build, parsed_toml):
        requires = build.get("requires", [])
        if not requires:
            return [], [], [], []

        library_names = []
        library_paths = []
        library_types = []
        for required in requires:
            the_dep = find_build(required, parsed_toml["builds"])
            library_types.append(the_dep["buildRule"])
            library_names.append(the_dep["outputName"])
            dep_name = the_dep["name"]
            library_paths.append(dep_name)

        library_paths = prepend_paths(build["build_dir"], library_paths)
        library_paths = PrefixLibraryPath(library_paths)
        return library_names, PrefixLibrary(library_names), library_paths, library_types

    def get_required_include_information(self, build, parsed_toml):
        requires = build.get("requires", [])
        if not requires:
            return []

        include_paths = []
        for required in requires:
            the_dep = find_build(required, parsed_toml["builds"])
            directory = build["directory"]
            includes = the_dep.get("includePaths", [])
            includes = prepend_paths(directory, includes)
            include_paths += includes

        return PrefixIncludePath(include_paths)

    def get_rpath(self, build: Dict, parsed_toml: Dict):
        # Good blog post about rpath:
        # https://medium.com/@nehckl0/creating-relocatable-linux-executables-by-setting-rpath-with-origin-45de573a2e98
        requires = build.get("requires", [])
        library_paths = []

        for required in requires:
            the_dep = find_build(required, parsed_toml["builds"])
            if the_dep["buildRule"] == "dynamiclib":
                library_paths.append(the_dep["name"])

        build_dir = Path(build["build_dir"]).resolve()
        current_build_dir = build_dir / build["name"]
        library_paths = prepend_paths(build_dir, library_paths)
        relative_paths = [
            relpath(Path(lib_path), current_build_dir) for lib_path in library_paths
        ]

        relative_paths = [f"$$ORIGIN/{rel_path}" for rel_path in relative_paths]
        relative_paths = ["$$ORIGIN"] + relative_paths

        relative_paths_string = escape_path(":".join(relative_paths))
        return f"-Wl,-rpath='{relative_paths_string}'"

    def add_naming_convention(self, output_name, build_type):
        if build_type == "staticlib":
            new_name = self.add_static_library_naming_convention(output_name)
        elif build_type == "dynamiclib":
            new_name = self.add_dynamic_library_naming_convention(output_name)
        else:
            new_name = self.add_exe_naming_convention(output_name)

        return new_name

    # TODO: These should take version strings as well.
    def add_static_library_naming_convention(self, library_name):
        return f"lib{library_name}.a"

    def add_dynamic_library_naming_convention(self, library_name):
        return f"lib{library_name}.so"

    def add_exe_naming_convention(self, exe_name):
        return f"{exe_name}.exe"


def log_build_information(build):
    build_name = build["name"]
    cxxflags = build["global_flags"] + build.get("flags", [])
    defines = build["global_defines"] + build.get("defines", [])
    includes = build["includes"]
    library_paths = build["libraryPaths"]
    output = build["outputName"]

    print(f"Running build: f{build_name}")
    print(f"FLAGS: {cxxflags}")
    print(f"DEFINES: {defines}")
    print(f"INCLUDE_PATHS: {includes}")
    print(f"LIBRARY_PATHS: {library_paths}")
    print(f"OUTPUT NAME: {output}")
    print("")



import argparse
import subprocess
import sys

import toml

from aim_build import gccbuilds
from aim_build import msvcbuilds
from aim_build import osxbuilds
from aim_build.schema import target_schema
from aim_build.utils import *
from aim_build.version import __version__


def find_build(build_name, builds):
    for build in builds:
        if build["name"] == build_name:
            return build
    else:
        raise RuntimeError(f"Failed to find build with name: {build_name}")


def run_ninja(working_dir, build_name):
    command = ["ninja", "-v", f"-C{build_name}", build_name]
    command_str = " ".join(command)
    print(f'Executing "{command_str}"')

    process = subprocess.Popen(
        command, cwd=str(working_dir), stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )

    for line in iter(process.stdout.readline, b""):
        sys.stdout.write(line.decode("utf-8"))
    for line in iter(process.stderr.readline, b""):
        sys.stderr.write(line.decode("utf-8"))


def run_ninja_generation(parsed_toml, project_dir: Path, build_dir: Path):
    compiler = parsed_toml["compiler"]
    archiver = parsed_toml["ar"]
    frontend = parsed_toml["compilerFrontend"]

    flags = parsed_toml.get("flags", [])
    defines = parsed_toml.get("defines", [])
    builds = parsed_toml["builds"]

    project_ninja = build_dir / "build.ninja"
    with project_ninja.open("w+") as project_fd:
        from ninja_syntax import Writer

        project_writer = Writer(project_fd)
        project_writer.include(str(build_dir / "rules.ninja"))

        for build_info in builds:
            print(f'Generating ninja file for {build_info["name"]}')
            build_info["directory"] = project_dir
            build_info["build_dir"] = build_dir
            build_info["global_flags"] = flags
            build_info["global_defines"] = defines
            build_info["global_compiler"] = compiler
            build_info["global_archiver"] = archiver

            if frontend == "msvc":
                # builder = msvcbuilds.MSVCBuilds(compiler, compiler_c, archiver)
                assert False, "MSVC frontend is currently not supported."
            elif frontend == "osx":
                builder = osxbuilds.OsxBuilds()
            else:
                builder = gccbuilds.GCCBuilds()

            builder.build(build_info, parsed_toml, project_writer)


def entry():
    # TODO: Get version automatically from the pyproject.toml file.
    parser = argparse.ArgumentParser(description=f"Version {__version__}")

    parser.add_argument("-v", "--version", action="version", version=__version__)
    sub_parser = parser.add_subparsers(dest="command", help="Commands")

    build_parser = sub_parser.add_parser(
        "list", help="displays the builds for the target"
    )
    build_parser.add_argument(
        "--target", type=str, required=True, help="path to target file directory"
    )

    init_parser = sub_parser.add_parser("init", help="creates a project structure")
    init_parser.add_argument(
        "--demo", help="Create additional demo files", action="store_true"
    )

    build_parser = sub_parser.add_parser("build", help="executes a build")
    build_parser.add_argument("build", type=str, help="The build name")

    build_parser.add_argument(
        "--target", type=str, required=True, help="path to target file directory"
    )

    build_parser.add_argument(
        "--skip_ninja_regen",
        help="by-pass the ninja file generation step",
        action="store_true",
    )

    build_parser = sub_parser.add_parser(
        "clobber", help="deletes all build artefacts for the specified target"
    )
    build_parser.add_argument(
        "--target", type=str, required=True, help="path to target file directory"
    )

    args = parser.parse_args()
    mode = args.command
    if mode == "init":
        run_init(args.demo)
    elif mode == "build":
        run_build(args.build, args.target, args.skip_ninja_regen)
    elif mode == "list":
        run_list(args.target)
    elif mode == "clobber":
        run_clobber(args.target)
    else:
        import sys

        parser.print_help(sys.stdout)


WindowsDefaultTomlFile = """\
cxx = "clang-cl"
cc = "clang-cl"
ar = "llvm-ar"
compilerFrontend="msvc"

flags = [
    "/std:c++17",
    "/Zi",
]

defines = []

#[[builds]]
#    name = "static"
#    buildRule = "staticlib"
#    outputName = "libraryName.lib"
#    srcDirs = ["../lib"]
#    includePaths = ["../include"]

[[builds]]
    name = "shared"
    buildRule = "dynamiclib"
    outputName = "libraryName.dll"
    srcDirs = ["../lib"]
    includePaths = ["../include"]

[[builds]]
    name = "exe"
    requires= ["shared"]
    buildRule = "exe"
    outputName = "exeName.exe"
    srcDirs = ["../app"]
    includePaths = ["../lib"]
    libraryPaths = ["./shared"]
    libraries = [""]
"""

LinuxDefaultTomlFile = """\
projectRoot = "../.."

cxx = "clang++"
cc = "clang"
ar = "ar"
compilerFrontend="gcc"

flags = [
    "-std=c++17",
    "-g"
]

defines = []

[[builds]]                              # a list of builds.
    name = "lib_calculator"             # the unique name for this build.
    buildRule = "staticlib"             # the type of build, in this case create a static library.
    outputName = "Calculator"      # the library output name,
    srcDirs = ["lib"]                   # the src directories  to build the static library from.
    includePaths = ["include"]    # additional include paths to use during the build.

#[[builds]]
#    name = "lib_calculator_so"         # the unique name for this build.
#    buildRule = "dynamiclib"           # the type of build, in this case create a shared library.
#    outputName = "Calculator"    # the library output name,
#    srcDirs = ["lib"]                  # the src directories to build the shared library from.
#    includePaths = ["include"]         # additional include paths to use during the build.

[[builds]]
    name = "exe"                        # the unique name for this build.
    buildRule = "exe"                   # the type of build, in this case an executable.
    requires = ["lib_calculator"]       # build dependencies. Aim figures out the linker flags for you.
    outputName = "the_calculator"   # the exe output name,
    srcDirs = ["src"]                   # the src directories to build the shared library from.
    includePaths = ["include"]          # additional include paths to use during the build.
    #libraryPaths = []                   # additional library paths, used for including third party libraries.
    #libraries = []                      # additional libraries, used for including third party libraries.
"""

CALCULATOR_CPP = """\
#include "calculator.h"

float add(float x, float y)
{
    return x + y;
}
"""

CALCULATOR_H = """\
#ifndef CALCULATOR_H
#define CALCULATOR_H

float add(float x, float y);

#endif
"""

MAIN_CPP = """\
#include "calculator.h"
#include <stdio.h>

int main()
{
    float result = add(40, 2);
    printf("The result is %f\\n", result);
    return 0;
}
"""


def run_init(add_demo_files):
    project_dir = Path().cwd()
    dirs = ["include", "src", "lib", "builds"]
    dirs = [project_dir / x for x in dirs]
    print(f"Creating directories...")
    for a_dir in dirs:
        print(f"\t{str(a_dir)}")
        a_dir.mkdir(exist_ok=True)

    windows_targets = [
        "windows-clang_cl-debug",
        "windows-clang_cl-release",
    ]

    linux_targets = ["linux-clang++-debug", "linux-clang++-release"]

    print("Creating common build targets...")
    build_dir = project_dir / "builds"
    for target in windows_targets:
        target_dir = build_dir / target
        target_dir.mkdir(exist_ok=True)
        # print(f"\t{str(target_dir)}")

        toml_file = target_dir / "target.toml"
        toml_file.touch(exist_ok=True)
        print(f"\t{str(toml_file)}")

        toml_file.write_text(WindowsDefaultTomlFile)

    for target in linux_targets:
        target_dir = build_dir / target
        target_dir.mkdir(exist_ok=True)
        # print(f"\t{str(target_dir)}")

        toml_file = target_dir / "target.toml"
        toml_file.touch(exist_ok=True)
        print(f"\t{str(toml_file)}")

        toml_file.write_text(LinuxDefaultTomlFile)

    if add_demo_files:
        (dirs[0] / "calculator.h").write_text(CALCULATOR_H)
        (dirs[1] / "main.cpp").write_text(MAIN_CPP)
        (dirs[2] / "calculator.cpp").write_text(CALCULATOR_CPP)


def run_build(build_name, target_path, skip_ninja_regen):
    print("Running build...")
    build_dir = Path().cwd()

    if target_path:
        target_path = Path(target_path)
        if target_path.is_absolute():
            build_dir = target_path
        else:
            build_dir = build_dir / Path(target_path)

    # ninja_path = project_dir / "build.ninja"
    toml_path = build_dir / "target.toml"

    with toml_path.open("r") as toml_file:
        parsed_toml = toml.loads(toml_file.read())

        builds = parsed_toml["builds"]
        the_build = find_build(build_name, builds)
        root_dir = parsed_toml["projectRoot"]
        project_dir = (build_dir / root_dir).resolve()
        assert project_dir.exists(), f"{str(project_dir)} does not exist."

        try:
            target_schema(parsed_toml, project_dir)
        except RuntimeError as e:
            print(f"Error: {e.args[0]}")
            exit(-1)

        if not skip_ninja_regen:
            print("Generating ninja files...")
            run_ninja_generation(parsed_toml, project_dir, build_dir)
            with (build_dir.resolve() / "compile_commands.json").open("w+") as cc:
                command = ["ninja", "-C", str(build_dir.resolve()), "-t", "compdb"]
                subprocess.run(command, stdout=cc)

        run_ninja(build_dir, the_build["name"])


def run_list(target_path):
    build_dir = Path().cwd()

    if target_path:
        target_path = Path(target_path)
        if target_path.is_absolute():
            build_dir = target_path
        else:
            build_dir = build_dir / Path(target_path)

    toml_path = build_dir / "target.toml"

    with toml_path.open("r") as toml_file:
        parsed_toml = toml.loads(toml_file.read())

        builds = parsed_toml["builds"]

        frontend = parsed_toml["compilerFrontend"]

        if frontend == "msvc":
            builder = msvcbuilds.MSVCBuilds("", "", "")
        elif frontend == "osx":
            builder = osxbuilds.OsxBuilds("", "", "")
        else:
            builder = gccbuilds.GCCBuilds("", "", "")

        header = ["Item", "Name", "Build Rule", "Output Name"]
        table = []

        for number, build in enumerate(builds):
            output_name = builder.add_naming_convention(
                build["outputName"], build["buildRule"]
            )
            row = [number, build["name"], build["buildRule"], output_name]
            table.append(row)

        from tabulate import tabulate

        print()
        print(tabulate(table, header))
        print()


def run_clobber(target_path):
    build_dir = Path().cwd()

    if target_path:
        target_path = Path(target_path)
        if target_path.is_absolute():
            build_dir = target_path
        else:
            build_dir = build_dir / Path(target_path)

    assert (build_dir / "target.toml").exists(), (
        f"Failed to find target.toml file in {str(build_dir)}.\n"
        "You might be trying to delete a directory that you want to keep."
    )

    print(f"Clobbering {str(build_dir)}...")

    dir_contents = build_dir.glob("*")
    for item in dir_contents:
        if item.name != "target.toml":
            print(f"Deleting {item.name}")
            if item.is_dir():
                import shutil

                shutil.rmtree(str(item))
            else:
                os.remove(str(item))


if __name__ == "__main__":
    entry()



from typing import Dict
from pathlib import Path
from aim_build.utils import prepend_paths, relpath, escape_path
from aim_build.gccbuilds import GCCBuilds
from aim_build.gccbuilds import find_build, PrefixLibraryPath, PrefixLibrary
from aim_build.osxbuildrules import *


def get_rpath(build: Dict, parsed_toml: Dict):
    # Good blog post about rpath:
    # https://medium.com/@nehckl0/creating-relocatable-linux-executables-by-setting-rpath-with-origin-45de573a2e98
    requires = build.get("requires", [])
    library_paths = []

    for required in requires:
        the_dep = find_build(required, parsed_toml["builds"])
        if the_dep["buildRule"] == "dynamiclib":
            library_paths.append(the_dep["name"])

    build_dir = Path(build["build_dir"]).resolve()
    current_build_dir = build_dir / build["name"]
    library_paths = prepend_paths(build_dir, library_paths)
    relative_paths = [
        relpath(Path(lib_path), current_build_dir) for lib_path in library_paths
    ]

    relative_paths = [f"@executable_path/{rel_path}" for rel_path in relative_paths]
    relative_paths = ["@executable_path"] + relative_paths

    relative_paths_string = escape_path(":".join(relative_paths))
    return f"-rpath '{relative_paths_string}'"


class OsxBuilds(GCCBuilds):
    def add_rules(self, build):
        directory = build["build_dir"]
        ninja_path = directory / "rules.ninja"
        with ninja_path.open("w+") as ninja_file:
            writer = Writer(ninja_file)
            add_compile(writer)
            add_ar(writer)
            add_exe(writer)
            add_shared(writer)

    def get_rpath(self, build: Dict, parsed_toml: Dict):
        return get_rpath(build, parsed_toml)

    # TODO: These should take version strings as well.
    def add_static_library_naming_convention(self, library_name):
        return f"lib{library_name}.a"

    def add_dynamic_library_naming_convention(self, library_name):
        return f"lib{library_name}.dylib"

    def add_exe_naming_convention(self, exe_name):
        return f"{exe_name}.exe"



from ninja_syntax import Writer


def add_compile(nfw: Writer):
    command = f"$compiler $defines $flags -MMD -MF deps.d $includes -c $in -o $out"
    nfw.rule(
        name="compile",
        description="Compiles source files into object files",
        deps="gcc",
        depfile="deps.d",
        command=command,
    )
    nfw.newline()


def add_ar(nfw: Writer):
    nfw.rule(
        name="archive",
        description="Combine object files into an archive",
        command="$archiver cr $out $in",
    )
    nfw.newline()


def add_exe(nfw: Writer):
    # TODO: origin should only really be added when we need to link against an so.
    command = f"$compiler $defines $flags $includes $in -o $exe_name $linker_args"
    nfw.rule(name="exe", description="Builds an executable.", command=command)
    nfw.newline()


def add_shared(nfw: Writer):
    command = f"$compiler $defines -shared -fvisibility=hidden -fPIC $flags $includes $in -o $out $linker_args"
    nfw.rule(name="shared", description="Builds a shared library.", command=command)
    nfw.newline()



import functools
from typing import Dict
from ninja_syntax import Writer

from aim_build.msvcbuildrules import *
from aim_build.utils import *

PrefixIncludePath = functools.partial(prefix, "/I")
PrefixLibraryPath = functools.partial(prefix, "/LIBPATH:")
PrefixLibrary = functools.partial(prefix, "")
ToObjectFiles = src_to_obj


def implicit_library_inputs(libraries):
    implicits = []
    for library in libraries:
        parts = library.split(".")
        if parts[1] == "dll":
            implicits.append(parts[0] + ".exp")
            implicits.append(parts[0] + ".lib")
        implicits.append(library)

    return implicits


def implicit_library_outputs(libraries):
    implicits = []
    for library in libraries:
        parts = library.split(".")
        if parts[1] == "dll":
            implicits.append(parts[0] + ".exp")
            implicits.append(parts[0] + ".lib")

    return implicits


def convert_dlls_to_lib(libraries):
    new_libraries = []
    for library in libraries:
        parts = library.split(".")
        if parts[1] == "dll":
            new_libraries.append(parts[0] + ".lib")
        else:
            new_libraries.append(library)

    return new_libraries


def get_src_files(build):
    directory = build["directory"]
    src_dirs = build["srcDirs"]
    src_paths = prepend_paths(directory, src_dirs)
    src_files = flatten(glob("*.cpp", src_paths)) + flatten(glob("*.c", src_paths))
    assert src_files, "Fail to find any source files."
    return src_files


def get_include_paths(build):
    directory = build["directory"]
    include_paths = build.get("includePaths", [])
    includes = prepend_paths(directory, include_paths)
    includes = PrefixIncludePath(add_quotes(includes))
    return includes


def get_library_paths(build):
    directory = build["directory"]
    library_paths = build.get("libraryPaths", [])
    library_paths = prepend_paths(directory, library_paths)
    library_paths = PrefixLibraryPath(add_quotes(library_paths))
    return library_paths


def get_library_information(build):
    libraries = build.get("libraries", [])
    implicits = implicit_library_inputs(libraries)
    link_libraries = PrefixLibrary(convert_dlls_to_lib(libraries))
    return libraries, implicits, link_libraries


def get_third_party_library_information(build):
    third_libraries = build.get("thirdPartyLibraries", [])
    third_libraries = PrefixLibrary(convert_dlls_to_lib(third_libraries))
    return third_libraries


class MSVCBuilds:
    def __init__(self, cxx_compiler, c_compiler, archiver):
        self.cxx_compiler = cxx_compiler
        self.c_compiler = c_compiler
        self.archiver = archiver

    def add_rules(self, build):
        directory = build["directory"]
        ninja_path = directory / "rules.ninja"
        with ninja_path.open("w+") as ninja_file:
            writer = Writer(ninja_file)
            add_compile(writer)
            add_ar(writer)
            add_exe(writer)
            add_shared(writer)

    def build(self, build):
        the_build = build["buildRule"]

        build_name = build["name"]
        project_dir = build["directory"]

        build_path = project_dir / build_name
        build_path.mkdir(parents=True, exist_ok=True)

        ninja_path = build_path / "build.ninja"
        build["buildPath"] = build_path

        self.add_rules(build)

        with ninja_path.open("w+") as ninja_file:
            ninja_writer = Writer(ninja_file)
            rule_path = (project_dir / "rules.ninja").resolve()
            ninja_writer.include(escape_path(str(rule_path)))
            ninja_writer.newline()

            if the_build == "staticlib":
                self.build_static_library(ninja_writer, build)
            elif the_build == "exe":
                self.build_executable(ninja_writer, build)
            elif the_build == "dynamiclib":
                self.build_dynamic_library(ninja_writer, build)
            else:
                raise RuntimeError(f"Unknown build type {the_build}.")

    def add_compile_rule(self, nfw: Writer, build: Dict):
        cxxflags = build["flags"]
        defines = build["defines"]

        src_files = get_src_files(build)
        includes = get_include_paths(build)

        # Its very important to specify the absolute path to the obj files.
        # This prevents recompilation of files when an exe links against a library.
        # Without the absolute path to the obj files, it would build the files again
        # in the current (exe's) build location.
        build_path = build["buildPath"]
        obj_files = ToObjectFiles(src_files)
        obj_files = prepend_paths(build_path, obj_files)

        file_pairs = zip(to_str(src_files), to_str(obj_files))
        for src_file, obj_file in file_pairs:
            nfw.build(
                outputs=obj_file,
                rule="compile",
                inputs=src_file,
                variables={
                    "compiler": self.cxx_compiler,
                    "includes": includes,
                    "flags": cxxflags,
                    "defines": defines,
                },
            )
            nfw.newline()

        return obj_files

    def build_static_library(self, nfw: Writer, build: Dict):
        build_name = build["name"]
        library_name = build["outputName"]

        obj_files = self.add_compile_rule(nfw, build)

        nfw.build(outputs=library_name, rule="ar", inputs=to_str(obj_files))
        nfw.newline()

        nfw.build(rule="phony", inputs=library_name, outputs=build_name)
        nfw.newline()

    def build_executable(self, nfw, build: Dict):
        build_name = build["name"]
        exe_name = build["outputName"]
        cxxflags = build["flags"]
        defines = build["defines"]
        requires = build.get("requires", [])
        build_path = build["buildPath"]

        includes = get_include_paths(build)
        library_paths = get_library_paths(build)
        libraries, implicits, link_libraries = get_library_information(build)
        third_libraries = get_third_party_library_information(build)

        linker_args = library_paths + link_libraries + third_libraries

        for requirement in requires:
            ninja_file = (build_path.parent / requirement / "build.ninja").resolve()
            assert ninja_file.exists(), f"Failed to find {str(ninja_file)}."
            nfw.subninja(escape_path(str(ninja_file)))
            nfw.newline()

        obj_files = self.add_compile_rule(nfw, build)

        nfw.build(
            outputs=exe_name,
            rule="exe",
            inputs=to_str(obj_files),
            implicit=implicits,
            variables={
                "compiler": self.cxx_compiler,
                "includes": includes,
                "flags": cxxflags,
                "defines": defines,
                "exe_name": exe_name,
                "linker_args": " ".join(linker_args),
            },
        )
        nfw.newline()

        nfw.build(rule="phony", inputs=exe_name, outputs=build_name)
        nfw.newline()

    def build_dynamic_library(self, nfw, build: Dict):
        build_name = build["name"]
        lib_name = build["outputName"]
        cxxflags = build["flags"]
        defines = build["defines"]

        includes = get_include_paths(build)
        library_paths = get_library_paths(build)
        libraries, implicits, link_libraries = get_library_information(build)
        third_libraries = get_third_party_library_information(build)

        linker_args = library_paths + link_libraries + third_libraries
        implicit_outputs = implicit_library_outputs([lib_name])

        obj_files = self.add_compile_rule(nfw, build)

        nfw.build(
            rule="shared",
            inputs=to_str(obj_files),
            outputs=lib_name,
            implicit=implicits,
            implicit_outputs=implicit_outputs,
            variables={
                "compiler": self.cxx_compiler,
                "includes": includes,
                "flags": " ".join(cxxflags),
                "defines": " ".join(defines),
                "lib_name": lib_name,
                "linker_args": " ".join(linker_args),
            },
        )
        nfw.newline()

        nfw.build(rule="phony", inputs=lib_name, outputs=build_name)
        nfw.newline()


def log_build_information(build):
    build_name = build["name"]
    cxxflags = build["flags"]
    defines = build["defines"]
    includes = build["includes"]
    library_paths = build["libraryPaths"]
    output = build["outputName"]

    print(f"Running build: f{build_name}")
    print(f"CXXFLAGS: {cxxflags}")
    print(f"DEFINES: {defines}")
    print(f"INCLUDE_PATHS: {includes}")
    print(f"LIBRARY_PATHS: {library_paths}")
    print(f"OUTPUT NAME: {output}")
    print("")
